/**************************************************************************
* document.cpp                                                            *
*                                                                         *
* Interface:  document.h                                                  *
* Implements: TDCDocument                                                 *
**************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <values.h>
#include "document.h"

/**************************************************************************
* fileerror TDCDocument::PhysicalLineBreak(int Count)                     *
*                                                                         *
* Puts Count physical linebreaks (0D0A) to the document.                  *
**************************************************************************/
fileerror TDCDocument::PhysicalLineBreak(int Count)
{
	for(int i=0;i<Count;i++)
		PutPhysical("\r\x0a");
	FOut->ResetCounter();
	return FOut->GetResult();
}

/**************************************************************************
* void TDCDocument::ResetNumberingCounts()                                *
*                                                                         *
* Resets the FNumberingCounts array, is to be called both                 *
* in the constructor and by BeginNumbering() when a completely new list   *
* is created.                                                             *
**************************************************************************/
void TDCDocument::ResetNumberingCounts()
{
	for(int i=0;i<MAX_BULLET_NUMBERING_LEVEL;i++)
		FNumberingCounts[i] = 1;
}

/**************************************************************************
* void TDCDocument::AllocateTableCellData()                               *
*                                                                         *
* Allocates pointers to longs for every item in the FCellData array.      *
* Called by PerformDefaultBeginTableAction().                             *
**************************************************************************/
void TDCDocument::AllocateTableCellData()
{
	for(int i=0;i<MAX_TABLE_COLUMNS;i++)
		for(int j=0;j<MAX_TABLE_ROWS;j++)
		{
			FCellData[i][j] = new long;
			*FCellData[i][j] = 0;
		}
}

/**************************************************************************
* void TDCDocument::FreeTableCellData()                                   *
*                                                                         *
* Frees the pointers allocated by AllocateTableCellData().                *
* Called by PerformDefaultBeginTableAction().                             *
**************************************************************************/
void TDCDocument::FreeTableCellData()
{
	for(int i=0;i<MAX_TABLE_COLUMNS;i++)
		for(int j=0;j<MAX_TABLE_ROWS;j++)
			delete FCellData[i][j];
}

/**************************************************************************
* void TDCDocument::SortLongArray(long a[], int size, sortorder order)    *
*                                                                         *
* Sorts the size size array a in using order as sort order.               *
**************************************************************************/
void TDCDocument::SortLongArray(long a[], int size, sortorder order)
{
	long extremevalue;
	int extremeindex;
	BOOL ready;
	for(int j=0;j<size;j++)
	{
		if(order == ASCEND)
			extremevalue = MAXLONG;
		else
			extremevalue = 0;
		extremeindex = 0;
		ready = TRUE;
		for(int i=j;i<size;i++)
		{
			if(((order == ASCEND) && (a[i] < extremevalue)) ||
				((order == DECEND) && (a[i] > extremevalue)))
			{
				ready = FALSE;
				extremeindex = i;
				extremevalue = a[i];
			}
		}
		if(!ready)
			SwapItemsInLongArray(a,j,extremeindex);
	}
}

/**************************************************************************
* void TDCDocument::SwapItemsInLongArray(long a[], int k, int l)	        *
*                                                                         *
* Swaps the two items with indecies k and l in the array a.               *
* No range checking is performed.                                         *
**************************************************************************/
void TDCDocument::SwapItemsInLongArray(long a[], int k, int l)
{
	long temp;
	temp = a[k];
	a[k] = a[l];
	a[l] = temp;
}

/**************************************************************************
* void TDCDocument::PerformDefaultBeginTableAction(char *TableCaption,    *
*	  captionalign ca)                                                     *
*                                                                         *
* Performs all default action that is common for all classes. Should be   *
* called by BeginTable() in every descendant. The parameters are the same *
* as BeginTable().                                                        *
**************************************************************************/
void TDCDocument::PerformDefaultBeginTableAction(char *TableCaption,
	captionalign ca)
{
	AllocateTableCellData();
	FIsTable = TRUE;
	FCurrentCol = -1;
	FCurrentRow = 0;
	FTableCaptionAlignment = ca;
	FTableCellWidth = 0;
	if(strlen(TableCaption))
	{
		FTableCaption = new char[strlen(TableCaption)+1];
		strcpy(FTableCaption,TableCaption);
	}
	else
		FTableCaption = NULL;
}

/**************************************************************************
* void TDCDocument::PerformDefaultEndTableAction()	                       *
*                                                                         *
* Performs all default action that is common for all classes. Should be   *
* called by EndTable() in every descendant.                               *
**************************************************************************/
void TDCDocument::PerformDefaultEndTableAction()
{
	FIsTable = FALSE;
	FTableOrdinal++;
	if(FTableCaption)
	{
		delete FTableCaption;
		FTableCaption = NULL;
	}
	FreeTableCellData();
}

/**************************************************************************
* fileerror TDCDocument::PutTableCaption()								        *
*                                                                         *
* Writes the already allocated table caption FTableCaption to             *
* with a label in the appropriate language. Normally called by            *
* BeginTable() and EndTable().                                            *
**************************************************************************/
fileerror TDCDocument::PutTableCaption()
{
	if(!FTableCaption)
		return OTHERERROR;
	switch(FLanguage)
	{
		case LANG_ENGLISH:	PutText("Table "); break;
		case LANG_SWEDISH:	PutText("Tabell "); break;
		case LANG_GERMAN:		PutText("Tabelle "); break;
		case LANG_FRENCH:		PutText("Table "); break;
		deafult:					PutText("Table "); break;
	}
	PutInt(FTableOrdinal);
	PutText(": ");
	PutText(FTableCaption);
	return FOut->GetResult();
}

/**************************************************************************
* void TDCDocument::BreakLine(char *dest, char *src)						     *
*                                                                         *
* Linebreaks up the src string and copies it into dest.                   *
**************************************************************************/
void TDCDocument::BreakLine(char *dest, char *src)
{
	strcpy(dest,src);
}

/**************************************************************************
* void TDCDocument::Translate(char *dest, char *src)							  *
*                                                                         *
* Takes src and performs any translations needed, e.g. HTML uses          *
* entities and RTF converts all characters > 127 to hexadecimal.          *
**************************************************************************/
void TDCDocument::Translate(char *dest, char *src)
{
	strcpy(dest,src);
}

/**************************************************************************
* BOOL TDCDocument::ActOnStyleSheet(int OldSS)							        *
*                                                                         *
* Inserts any tags and/or line breaks when a new style sheet is selected. *
**************************************************************************/
BOOL TDCDocument::ActOnStyleSheet(int OldSS)
{
	return TRUE;
}

BOOL TDCDocument::SetBold(BOOL Toggle)
{
	return FALSE;
}

BOOL TDCDocument::SetItalic(BOOL Toggle)
{
	return FALSE;
}

ulstyle TDCDocument::SetUnderline(ulstyle Style)
{
	return NONE;
}

BOOL TDCDocument::SetStrikethrough(BOOL Toggle)
{
	return FALSE;
}

BOOL TDCDocument::SetHidden(BOOL Toggle)
{
	return FALSE;
}

/**************************************************************************
* int TDCDocument::SetStyleSheet(int NewSS)									     *
*                                                                         *
* Sets the style sheet to the one specified by NewSS and makes            *
* a call to ActOnStyleSheet().                                            *
**************************************************************************/
int TDCDocument::SetStyleSheet(int NewSS)
{
	int temp;
	temp = FStyleSheet;
	FStyleSheet = NewSS;
	ActOnStyleSheet(temp);
	return temp;
}

char *TDCDocument::SetAuthor(char *NewAuthor)
{
	char *temp = "\0";
	if(FAuthor)
	{
		strcpy(temp,FAuthor);
		delete FAuthor;
	}
	else
		temp = "";
	FAuthor = new char[strlen(NewAuthor)+1];
	strcpy(FAuthor,NewAuthor);
	return temp;
}

char *TDCDocument::SetTitle(char* NewTitle)
{
	char *temp = "\0";
	if(FTitle)
	{
		strcpy(temp,FTitle);
		delete FTitle;
	}
	else
		temp = "";
	FTitle = new char[strlen(NewTitle)+1];
	strcpy(FTitle,NewTitle);
	return temp;
}

char *TDCDocument::SetKeywords(char *NewKeywords)
{
	char *temp = "\0";
	if(FKeywords)
	{
		strcpy(temp,FKeywords);
		delete FKeywords;
	}
	FKeywords = new char[strlen(NewKeywords)+1];
	strcpy(FKeywords,NewKeywords);
	return temp;
}

char TDCDocument::SetDividerChar(char NewDivider)
{
	char temp = FDividerChar;
	FDividerChar = NewDivider;
	return temp;
}

int TDCDocument::SetLineWidth(int NewWidth)
{
	int temp = FLineLength;
	FLineLength = NewWidth;
	return temp;
}

BOOL TDCDocument::SetWordWrap(BOOL NewValue)
{
	BOOL temp = FWordWrap;
	FWordWrap = NewValue;
	if(FWordWrap)
		FOut->StartCounter();
	else
	{
		FOut->ResetCounter();
		FOut->StopCounter();
	}
	return temp;
}

BOOL TDCDocument::GetBold()
{
	return FBold;
}

BOOL TDCDocument::GetItalic()
{
	return FItalic;
}

ulstyle TDCDocument::GetUnderline()
{
	return FUnderline;
}

BOOL TDCDocument::GetStrikethrough()
{
	return FStrikethrough;
}

BOOL TDCDocument::GetHidden()
{
	return FHidden;
}

int TDCDocument::GetStyleSheet()
{
	return FStyleSheet;
}

char *TDCDocument::GetAuthor()
{
	return FAuthor;
}

char *TDCDocument::GetTitle()
{
	return FTitle;
}

char *TDCDocument::GetExtension()
{
	return FExtension;
}

char *TDCDocument::GetFilename()
{
	return FOut->GetFilename();
}

char *TDCDocument::GetKeywords()
{
	return FKeywords;
}

char TDCDocument::GetDelimiterChar()
{
	return FDividerChar;
}

charset TDCDocument::GetCharset()
{
	return FCharset;
}

int TDCDocument::GetLineWidth()
{
	return FLineLength;
}

BOOL TDCDocument::GetWordWrap()
{
	return FWordWrap;
}

int TDCDocument::GetBulletLevel()
{
	return FBulletLevel;
}

int TDCDocument::GetNumberingLevel()
{
	return FNumberingLevel;
}

/**************************************************************************
* fileerror TDCDocument::BeginBullets()										     *
*                                                                         *
* Begins a new bulleted list. Must increase FBulletLevel                  *
* and set FIsFirstBullet to TRUE.                                         *
**************************************************************************/
fileerror TDCDocument::BeginBullets()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::EndBullets()										        *
*                                                                         *
* Ends a bulleted list (make sure one has been started). Must decrease    *
* FBulletLevel and set FIsFirstBullet and FIsFirstNumber to FALSE.        *
**************************************************************************/
fileerror TDCDocument::EndBullets()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::NewBullet()                                      *
*                                                                         *
* Inserts a new bullet in a bulleted list. If FIsFirstBullet is set, it   *
* should be nulled.                                                       *
**************************************************************************/
fileerror TDCDocument::NewBullet()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::BeginNumbering()										     *
*                                                                         *
* Begins a new bulleted list. Must increase FNumberingLevel, the          *
* appropriate index (FNumberingLevel) of FNumberingCounts                 *
* and set FIsFirstBullet to TRUE.                                         *
**************************************************************************/
fileerror TDCDocument::BeginNumbering()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::EndNumbering()										     *
*                                                                         *
* End a new bulleted list. Must decrease FNumberingLevel.                 *
**************************************************************************/
fileerror TDCDocument::EndNumbering()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::NewNumber()                                      *
*                                                                         *
* Inserts a new number in a bulleted list. If FIsFirstNumber is set, it   *
* should be nulled.                                                       *
**************************************************************************/
fileerror TDCDocument::NewNumber()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::BeginTable(char *TableCaption, captionalign ca)  *
*                                                                         *
* Begins a new table with TableCaption as table caption and the caption   *
* aligned to ca. Calls PerformDefaultBeginTableAction and sets            *
* FAccumulatedTableWidth to zero.                                         *
**************************************************************************/
fileerror TDCDocument::BeginTable(char *TableCaption, captionalign ca)
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::EndTable()                                       *
*                                                                         *
* Begins a new table with TableCaption as table caption and the caption   *
* aligned to ca. Calls PerformDefaultBeginTableAction and sets            *
* FAccumulatedTableWidth to zero.                                         *
**************************************************************************/
fileerror TDCDocument::EndTable()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::AddCell(int Width, int BorderStyle)              *
*                                                                         *
* Adds a cells to the table. The user is supposed to do this even when    *
* the table is empty or a new line has been inserted. This behaviour must *
* be supported. FCurrentCol must be increased.                            *
**************************************************************************/
fileerror TDCDocument::AddCell(int Width, int BorderStyle)
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::AddRow()                                         *
*                                                                         *
* Adds a new empty row to the table. The user is supposed to do this even *
* when the table is empty or a new line has been inserted. This behaviour *
* must be supported. FCurrentRow must also be increased and FCurrentColl  *
* is to be set to -1.                                                     *
**************************************************************************/
fileerror TDCDocument::AddRow()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::InsertFormula(int Range, int Formula, int Skip)  *
*                                                                         *
* Insert the value calculated from the table cell data in FCellData. Uses *
* the formula specified in Formula (sum, mean or median) and calculates   *
* from the cell range Range. Skip tells how many columns/rows at the      *
* side/top that are to be excluded form the calculation. Since the cell   *
* data should be set to zero if the cell value isn't a valid number, Skip *
* could usually be 0.                                                     *
**************************************************************************/
fileerror TDCDocument::InsertFormula(int Range, int Formula, int Skip)
{
	long result = 0;
	int n = 0;
	if((Formula == CF_SUM) || (Formula == CF_MEAN))
	{
		if(Range == CR_ABOVE)
		{
			for(int i=Skip;i<FCurrentRow;i++)
			{
				if(*FCellData[FCurrentCol][i])
				{
					result += *FCellData[FCurrentCol][i];
					n++;
				}
			}
		}
		else if (Range == CR_LEFT)
		{
			for(int i=Skip;i<FCurrentCol;i++)
			{
				if(*FCellData[i][FCurrentRow])
				{
					result += *FCellData[i][FCurrentRow];
					n++;
				}
			}
		}
		if(Formula == CF_MEAN)
			result /= n;
	}
	if(Formula == CF_MEDIAN)
	{
		long sorted[MAX_TABLE_ROWS];
		int pos = 0;
		if(Range == CR_ABOVE)
		{
			for(int i=Skip;i<FCurrentRow;i++)
			{
				if(*FCellData[FCurrentCol][i])
				{
					sorted[pos] = *FCellData[FCurrentCol][i];
					pos++;
				}
			}
		}
		else if(Range == CR_LEFT)
		{
			for(int i=Skip;i<FCurrentRow;i++)
			{
				if(*FCellData[i][FCurrentRow])
				{
					sorted[pos] = *FCellData[i][FCurrentRow];
					pos++;
				}
			}
		}
		if(!pos)
			result = 0;
		else
		{
			SortLongArray(sorted,pos);
			if(pos == 1)
				result = sorted[0];
			else if(pos == 2)
				result = (sorted[0] + sorted[1]) / 2;
			else if(pos % 2)					// Is the number odd?
			{
				int i = pos / 2;
				result = sorted[i];
			}
			else
			{
				int i = (pos / 2) - 1;
				int j = pos / 2;
				result = (sorted[i] + sorted[j]) / 2;
			}
		}
	}
	PutLong(result);
	return FOut->GetResult();
}

/**************************************************************************
* fileerror TDCDocument::PutText(char *text)                              *
*                                                                         *
* Writes the text text to the document, linebreaking and translating as   *
* necessary. Must convert text into a long and assign it to               *
* FCellData[FCurrentCol][FCurrentRow].                                    *
**************************************************************************/
fileerror TDCDocument::PutText(char *text)
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::PutInt(int Num)										     *
*                                                                         *
* Writes an integer value Num to the document. Calls PutText() after Num  *
* has been converted to a string. Therefore, don't worry about FCellData.       *
**************************************************************************/
fileerror TDCDocument::PutInt(int Num)
{
	char temp[13];
	itoa(Num,temp,10);
	PutText("Trams");
	return FOut->GetResult();
}

/**************************************************************************
* fileerror TDCDocument::PutLong(long Num)										  *
*                                                                         *
* Writes a long integer value Num to the document. Calls PutText()        *
* after Num has been converted to a string. Therefore, don't worry        *
* about FCellData.                                                        *
**************************************************************************/
fileerror TDCDocument::PutLong(long Num)
{
	char *temp = new char[13];
	ltoa(Num,temp,10);
	PutText(temp);
	delete temp;
	return FOut->GetResult();
}

/**************************************************************************
* fileerror TDCDocument::PutPhysical(char *Text, unsigned Length)			  *
*                                                                         *
* Writes Text with length Length to the document without performing any   *
* kind of translation. Use PutPhysical() only if you're putting tags, use *
* PutText() otherwise. *Writes a long integer value Num to the document.  *
* Calls PutText() after Num has been converted to a string. Therefore,    *
* don't worry about FCellData.                                            *
**************************************************************************/
fileerror TDCDocument::PutPhysical(char *Text, unsigned Length)
{
	if(Length)
		return FOut->Write(Text,Length);
	else
		return FOut->Write(Text,strlen(Text));
}

/**************************************************************************
* fileerror TDCDocument::PutPhysical(char *Text1, char *Text2,            *
*    char *Text3)                                                         *
*                                                                         *
* Same as PutPhysical(char *, unsigned) except that it takes three        *
* strings (only the first is mandatory.                                   *
**************************************************************************/
fileerror TDCDocument::PutPhysical(char *Text1, char *Text2, char *Text3)
{
	PutPhysical(Text1,strlen(Text1));
	if(strlen(Text2))
		PutPhysical(Text2,strlen(Text2));
	if(strlen(Text3))
		PutPhysical(Text3,strlen(Text3));
	return FOut->GetResult();
}

/**************************************************************************
* fileerror TDCDocument::PutPhysical(int Num)								     *
*                                                                         *
* Converts Num to a string literal and writes it to FOut.                 *
**************************************************************************/
fileerror TDCDocument::PutPhysical(int Num)
{
	char *temp = new char[10];
	itoa(Num,temp,10);
	PutPhysical(temp);
	delete temp;
	return FOut->GetResult();
}

/**************************************************************************
* fileerror TDCDocument::PutHeaders()            							     *
*                                                                         *
* Puts all necessary headers at the beginning of the document. Should be  *
* called before all 'put' operations, but after setting all document      *
* variables with SetTitle(), SetAuthor() etc.                             *
**************************************************************************/
fileerror TDCDocument::PutHeaders()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::PutFooters()            							     *
*                                                                         *
* Puts all necessary headers at the end of the document. This call should *
* the last thing done before Close() is called.                           *
**************************************************************************/
fileerror TDCDocument::PutFooters()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::PutDivider()            							     *
*                                                                         *
* Puts a document divider, usually a horizontal line                      *
* across the full width                                                   *
**************************************************************************/
fileerror TDCDocument::PutDivider()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::PutLink(char *LinkText, char *LinkTo)			     *
*                                                                         *
* Inserts a hypertext link into the document. LinkText is the displayed   *
* text and LinkTo specifies the location of the destination and is indi-  *
* vidual for all document types. Since hypertext links not always are     *
* possible, PutLink() defaults to simply putting LinkText.                *
**************************************************************************/
fileerror TDCDocument::PutLink(char *LinkText, char *LinkTo)
{
	return PutText(LinkText);
}

/**************************************************************************
* fileerror TDCDocument::ParagraphBreak()                     			     *
*                                                                         *
* Inserts a paragraph break.                                              *
**************************************************************************/
fileerror TDCDocument::ParagraphBreak()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::LineBreak()                          			     *
*                                                                         *
* Inserts a line break.                                                   *
**************************************************************************/
fileerror TDCDocument::LineBreak()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::Tab()                          			           *
*                                                                         *
* Inserts a tab.                                                          *
**************************************************************************/
fileerror TDCDocument::Tab()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::LineBreak()                          			     *
*                                                                         *
* Inserts a deliberate line break.                                        *
**************************************************************************/
fileerror TDCDocument::PageBreak()
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::NewTopic(char *TopicTitle, char *TopicName)      *
*                                                                         *
* Creates a new topic with the specified title and name (an internal name *
* of the topic). The method should automatically switch style sheet as    *
* necessary.                                                              *
**************************************************************************/
fileerror TDCDocument::NewTopic(char *TopicTitle, char *TopicName)
{
	return OK;
}

/**************************************************************************
* fileerror TDCDocument::Open(BOOL append)                                *
*                                                                         *
* Tells FOut to open whatever media the destination may be and prepare it *
* for writing. If append is true, the output should be appended to        *
* existing data, if any.                                                  *
**************************************************************************/
fileerror TDCDocument::Open(BOOL append)
{
	return FOut->Open(append);
}

/**************************************************************************
* fileerror TDCDocument::Close()                                          *
*                                                                         *
* Tells FOut to close whatever media the destination may be and prepare   *
* for shutdown.                                                           *
**************************************************************************/
fileerror TDCDocument::Close()
{
	return FOut->Close();
}

/**************************************************************************
* int TDCDocument::GetIdent()                                             *
*                                                                         *
* Should return a unique identification constant, preferably              *
* prefixed by DT_.                                                        *
**************************************************************************/
int TDCDocument::GetIdent()
{
	return 0;
}

/**************************************************************************
* fileerror TDCDocument::GetError()                                       *
*                                                                         *
* Returns the current error state of FOut.                                *
**************************************************************************/
fileerror TDCDocument::GetError()
{
	if(FOut)
		return FOut->GetResult();
	else
		return NOTOPEN;
}

TDCDocument::TDCDocument(char *Filename, TOutput *OutputObj)
{
	FOut = OutputObj;
	FOut->SetFilename(Filename);
	FExtension = new char[4];
	FAuthor = NULL;
	FTitle = NULL;
	FKeywords = NULL;
	FTableCaption = NULL;
	FLineLength = 80;
	FWordWrap = TRUE;
	FBold = FALSE;
	FItalic = FALSE;
	FUnderline = NONE;
	FStrikethrough = FALSE;
	FHidden = FALSE;
	FStyleSheet = SS_NORMAL;
	FBulletLevel = 0;
	FNumberingLevel = 0;
	FCurrentRow = 0;
	FCurrentCol = 0;
	FTableOrdinal = 1;
	FLanguage = LANG_ENGLISH;
	FDividerChar = '=';
	FIsFirstLineOfDoc = TRUE;
	FIsTable = FALSE;
	if(FWordWrap)
		FOut->StartCounter();
	ResetNumberingCounts();
}

TDCDocument::~TDCDocument()
{
	if(FOut)
		delete FOut;
	if(FAuthor)
		delete FAuthor;
	if(FTitle)
		delete FTitle;
	if(FExtension)
		delete FExtension;
	if(FKeywords)
		delete FKeywords;
}
