#ifndef __DOCUMENT_H

#include <stdio.h>
#include "basicdef.h"
#include "foutstd.h"

typedef enum charset
{
	UNKNOWN,
	PCWINDOWS,
	PCDOS,
	UNIX,
	MAC,
	ISO
};

typedef enum ulstyle
{
	NONE,
	SINGLE,
	DOUBLE,
	BYWORD
};

// Where the table captions are to be placed
typedef enum captionalign
{
	ABOVE,
	BELOW
};

// Sort order to be used in SortLongArray()
typedef enum sortorder
{
	ASCEND,			// Ascending order, 1-2-3
	DECEND			// Descending order, 3-2-1
};

// Stylesheets
const int SS_NORMAL			= 0;
const int SS_FIRSTLINE     = 1;
const int SS_HEADING1      = 2;
const int SS_HEADING2      = 3;
const int SS_HEADING3      = 4;
const int SS_CAPTION		   = 5;

// The language currently used in document
const int LANG_UNKNOWN		= 0;
// These constants are predefined in Win32
#ifndef __WIN32__
const int LANG_ENGLISH		= 0x09;
const int LANG_SWEDISH		= 0x1d;
const int LANG_GERMAN		= 0x07;
const int LANG_FRENCH		= 0x0c;
#endif

// Number of characters to indent the first line with in a text document
const int FIRST_LINE_INDENT				= 3;

// Bullet and numbering-related constants
const int MAX_BULLET_NUMBERING_LEVEL 	= 9;
const int BULLET_NUMBERING_INDENT		= 2;
const int HANGING_INDENT					= 4;

// Color constants
const int COLOR_BLACK					= 0;
const int COLOR_LIGHT_BLUE 	  		= 1;
const int COLOR_LIGHT_CYAN				= 2;
const int COLOR_LIGHT_GREEN	  		= 3;
const int COLOR_LIGHT_MAGENTA			= 4;
const int COLOR_BLUE						= 5;
const int COLOR_CYAN						= 6;
const int COLOR_GREEN					= 7;
const int COLOR_MAGENTA					= 8;
const int COLOR_RED						= 9;
const int COLOR_YELLOW					= 10;
const int COLOR_GRAY						= 11;
const int COLOR_LIGHT_GRAY				= 12;
const int COLOR_LIGHT_RED				= 13;
const int COLOR_LIGHT_YELLOW			= 14;
const int COLOR_WHITE					= 15;

// The size of a buffer used for allocation of some strings
const int BUF_SIZE			  	= 2048;

// The maximum size of a calculated table
const int MAX_TABLE_COLUMNS	= 16;
const int MAX_TABLE_ROWS		= 512;

// Table calculation ranges
const int CR_ABOVE				= 1;
const int CR_LEFT					= 2;

// Table calculation formulae
const int CF_SUM					= 1;
const int CF_MEAN					= 2;
const int CF_MEDIAN				= 3;

// Border styles
const int BS_NONE					= 0;
const int BS_TOP					= 1;
const int BS_LEFT					= 2;
const int BS_RIGHT				= 4;
const int BS_BOTTOM				= 8;

class TDCDocument
{
	protected:
//*** Protected data members
		BOOL FBold;
		BOOL FItalic;
		BOOL FStrikethrough;
		BOOL FHidden;
		BOOL FWordWrap;
		BOOL FIsFirstLineOfDoc;		// Is the current line the first of this doc?
		BOOL FIsTable;					// Is a table being created?
		ulstyle FUnderline;
		char *FAuthor;
		char *FTitle;
		char *FExtension;				// The file extension
		char *FKeywords;
		char *FTableCaption;			// The caption of the current table
		char FLastWritten;        	// The last character written
		char FDividerChar;			// The character to use for text
											// dividers, defaults to '='
		unsigned char FLineLength; // The max length of a line, defaults to 80
		int FLanguage;            	// Document language, defaults to English
		int FBulletLevel;         	// Current bullet level, 0 if no bullets
		int FNumberingLevel;      	// Current numbering level, 0 if no numbering
		int FCurrentCol;
		int FCurrentRow;
		int FTableOrdinal;
		int FTableCellWidth;			// The requested width of the current cell
		int FStyleSheet;
		captionalign FTableCaptionAlignment;
		long *FCellData[MAX_TABLE_COLUMNS][MAX_TABLE_ROWS];
		TOutput *FOut;
		charset FCharset;
//*** Allow 10 levels of numbers
		int FNumberingCounts[MAX_BULLET_NUMBERING_LEVEL];
//*** Protected member functions
		fileerror PhysicalLineBreak(int Count = 1);
		void ResetNumberingCounts();
		void AllocateTableCellData();
		void FreeTableCellData();
		void SortLongArray(long array[],int size, sortorder order = ASCEND);
		void SwapItemsInLongArray(long a[], int k, int l);
		void PerformDefaultBeginTableAction(char *TableCaption, captionalign ca);
		void PerformDefaultEndTableAction();
		fileerror PutTableCaption();
		virtual void BreakLine(char *dest, char *src);
		virtual void Translate(char *dest, char *src);
		virtual BOOL ActOnStyleSheet(int OldSS);
	public:
//*** Set text attributes
		virtual BOOL SetBold(BOOL Toggle);
		virtual BOOL SetItalic(BOOL Toggle);
		virtual ulstyle SetUnderline(ulstyle Style);
		virtual BOOL SetStrikethrough(BOOL Toggle);
		virtual BOOL SetHidden(BOOL Toggle);
		virtual int SetStyleSheet(int NewSS);
//*** Get text attributes
		virtual BOOL GetBold();
		virtual BOOL GetItalic();
		virtual ulstyle GetUnderline();
		virtual BOOL GetStrikethrough();
		virtual BOOL GetHidden();
		virtual int GetStyleSheet();
//*** Set document attributes
		char *SetAuthor(char *NewAuthor);
		char *SetTitle(char* NewTitle);
		char *SetKeywords(char *NewKeywords);
		char SetDividerChar(char NewDelimiter);
		int SetLineWidth(int NewWidth);
		BOOL SetWordWrap(BOOL NewValue);
//*** Get document attributes
		char *GetAuthor();
		char *GetTitle();
		char *GetExtension();
		char *GetFilename();
		char *GetKeywords();
		char GetDelimiterChar();
		charset GetCharset();
		int GetLineWidth();
		BOOL GetWordWrap();
//*** Bullet and numbering functions
		int GetBulletLevel();
		int GetNumberingLevel();
		virtual fileerror BeginBullets();
		virtual fileerror EndBullets();
		virtual fileerror NewBullet();
		virtual fileerror BeginNumbering();
		virtual fileerror EndNumbering();
		virtual fileerror NewNumber();
//*** Table functions
		virtual fileerror BeginTable(char *TableCaption = "",
			captionalign ca = BELOW);
		virtual fileerror EndTable();
		virtual fileerror AddCell(int Width = 0, int BorderStyle = BS_NONE);
		virtual fileerror AddRow();
		virtual fileerror InsertFormula(int Range, int Formula, int Skip = 0);
//*** Write to the document
		virtual fileerror PutText(char *Text);
		virtual fileerror PutInt(int Num);
		virtual fileerror PutLong(long Num);
		virtual fileerror PutPhysical(char *Text, unsigned Length);
		virtual fileerror PutPhysical(char *Text1, char *Text2 = "",
			char *Text3 = "");
		virtual fileerror PutPhysical(int Num);
		virtual fileerror PutHeaders();
		virtual fileerror PutFooters();
		virtual fileerror PutDivider();
		virtual fileerror Tab();
		virtual fileerror PutLink(char *LinkText, char *LinkTo);
		virtual fileerror ParagraphBreak();
		virtual fileerror LineBreak();
		virtual fileerror PageBreak();
		virtual fileerror NewTopic(char *TopicTitle, char *TopicName);
//*** Close and open file
		virtual fileerror Open(BOOL Append = FALSE);
		virtual fileerror Close();
//*** Misc methods
		virtual int GetIdent();
		fileerror GetError();
//*** Public constructors and destructors
		TDCDocument(char *Filename, TOutput *OutputObj);
		~TDCDocument();
};

#define __DOCUMENT_H

#endif
