/**************************************************************************
* textdoc.cpp                                                             *
*                                                                         *
* Interface:  textdoc.h                                                   *
* Implements: TDCTextDocument                                             *
*                                                                         *
* See also document.cpp for additional comments of common methods.        *
**************************************************************************/

#include <string.h>
#include <cstring.h>
#include "textdoc.h"

/**************************************************************************
* void TDCTextDocument::BreakLine(char *dest, char *src)                  *
*                                                                         *
* Performs line breaking. Since complete lines don't have to be put with  *
* PutText(), this is a little tricky. The counter of FOut is increased by *
* the used TOutput descendant with the length of src. If the resulting    *
* string is longer than FLineLength, the best place to place the line     *
* break is found by finding the last space that's within FLineLength.     *
* There, a line break is placed and the counter is reset.                 *
**************************************************************************/
void TDCTextDocument::BreakLine(char *dest, char *src)
{
	string *s = new string(src);
	int i, currcount, lastspace = 0;
	size_t temp;
	currcount = FOut->GetCounter();
	if((currcount + s->length()) > FLineLength)
	{
		for(i=0;i<(s->length()-1);i+=FLineLength)
		{
			lastspace = 0;
			do
			{
				temp = s->find(" ",(i * FLineLength) + lastspace + 1);
				if((temp != NPOS) && (temp < (FLineLength - currcount - 1)))
					lastspace = temp;
			} while ((temp != NPOS) && (temp < (FLineLength - currcount - 1)));
			if(lastspace)
			{
				FIsFirstLineOfDoc = FALSE;
				s->insert(lastspace+1,"\r\n");
				if(FBulletLevel || FNumberingLevel)
				{
					char blanks[BULLET_NUMBERING_INDENT * 10];
					for(int i=0;i<(BULLET_NUMBERING_INDENT * 10) - 1;i++)
						blanks[i] = ' ';
					blanks[BULLET_NUMBERING_INDENT * (FNumberingLevel +
						FBulletLevel)+4] = '\0';
					s->insert(lastspace+3,blanks);
				}
				FOut->ResetCounter();
			}
		}
	}
	strcpy(dest,s->c_str());
	delete s;
}

/**************************************************************************
* void TDCTextDocument::BreakLine(char *dest, char *src)                  *
*                                                                         *
* Used by the table methods to insert spaces to pad a table cell to the   *
* cell width specified in FTableCellWidth.                                *
**************************************************************************/
fileerror TDCTextDocument::PadCellToRequestedLength()
{
	char temp[255] = "";
	if((!FTableCellWidth) || (FCellLength > FTableCellWidth))
		return OK;
	for(int i=0;i<(FTableCellWidth - FCellLength);i++)
		strcat(temp," ");
	return PutText(temp);
}

BOOL TDCTextDocument::ActOnStyleSheet(int OldSS)
{
	if((OldSS != SS_NORMAL) && (OldSS != SS_FIRSTLINE))
	{
		PhysicalLineBreak();
		if((!FIsFirstLineOfDoc) && (FOut->GetLastChar() != '\n'))
			PhysicalLineBreak();
	}
	if((!FIsFirstLineOfDoc) && FStyleSheet != SS_NORMAL)
		PhysicalLineBreak();
	FIsFirstLineOfDoc = FALSE;
	if(FStyleSheet == SS_FIRSTLINE)
	{
		for(int i=0;i<FIRST_LINE_INDENT;i++)
			PutText(" ");
	}
	return FOut->GetResult();
}

fileerror TDCTextDocument::BeginBullets()
{
	if(FBulletLevel < MAX_BULLET_NUMBERING_LEVEL)
		FBulletLevel++;
	return OK;
}

fileerror TDCTextDocument::EndBullets()
{
	if(FBulletLevel > 0)
		FBulletLevel--;
	PhysicalLineBreak();
	PhysicalLineBreak();
	return FOut->GetResult();
}

fileerror TDCTextDocument::NewBullet()
{
	char blanks[BULLET_NUMBERING_INDENT * 10];
	for(int i=0;i<BULLET_NUMBERING_INDENT * 10 - 1;i++)
		blanks[i] = ' ';
	blanks[BULLET_NUMBERING_INDENT * (FNumberingLevel + FBulletLevel)] = '\0';
	if(FOut->GetLastChar() != '\n')
	{
		PhysicalLineBreak();
		PhysicalLineBreak();
	}
	PutPhysical(blanks);
	PutPhysical("*   ");
	return FOut->GetResult();
}

fileerror TDCTextDocument::BeginNumbering()
{
	if(FNumberingLevel < MAX_BULLET_NUMBERING_LEVEL)
	{
		FNumberingLevel++;
		FNumberingCounts[FNumberingLevel] = 1;
	}
	return OK;
}

fileerror TDCTextDocument::EndNumbering()
{
	if(FNumberingLevel > 0)
		FNumberingLevel--;
	PhysicalLineBreak();
	PhysicalLineBreak();
	return FOut->GetResult();
}

fileerror TDCTextDocument::NewNumber()
{
	char blanks[BULLET_NUMBERING_INDENT * 10];
	for(int i=0;i<BULLET_NUMBERING_INDENT * 10 - 1;i++)
		blanks[i] = ' ';
	blanks[BULLET_NUMBERING_INDENT * (FNumberingLevel + FBulletLevel)] = '\0';
	if(FOut->GetLastChar() != '\n')
		PhysicalLineBreak(2);
	PutPhysical(blanks);
	char s[4] = "";
	itoa(FNumberingCounts[FNumberingLevel],s,10);
	strcat(s,".");
	while (strlen(s) < 4)
		strcat(s," ");
	PutPhysical(s);
	FNumberingCounts[FNumberingLevel]++;
	return FOut->GetResult();
}

fileerror TDCTextDocument::BeginTable(char *TableCaption, captionalign ca)
{
	PerformDefaultBeginTableAction(TableCaption,ca);
	FCellLength = 0;
	if(FTableCaption && (FTableCaptionAlignment == ABOVE))
	{
		if(FOut->GetLastChar() != '\n')
			LineBreak();
		PutTableCaption();
		LineBreak();
	}
	return OK;
}

fileerror TDCTextDocument::EndTable()
{
	LineBreak();
	if(FTableCaption && (FTableCaptionAlignment == BELOW))
	{
		PutTableCaption();
		LineBreak();
	}
	PerformDefaultEndTableAction();
	return OK;
}

fileerror TDCTextDocument::AddCell(int Width, int BorderStyle)
{
	FCurrentCol++;
	if(FCurrentCol == 0)
	{
		FTableCellWidth = Width;
		FCellLength = 0;
		return OK;
	}
	if(FTableCellWidth)
	{
		PadCellToRequestedLength();
		FTableCellWidth = Width;
		FCellLength = 0;
		return OK;
	}
	else
	{
		FTableCellWidth = Width;
		FCellLength = 0;
		return Tab();
	}
}

fileerror TDCTextDocument::AddRow()
{
	FCurrentRow++;
	FCurrentCol = -1;
	return LineBreak();
}

fileerror TDCTextDocument::PutText(char *text)
{
	char *temp = new char[BUF_SIZE];
	BreakLine(temp,text);
	*FCellData[FCurrentCol][FCurrentRow] = atol(temp);
	FOut->Write(temp,strlen(temp));
	FCellLength += strlen(temp);
	delete temp;
	return FOut->GetResult();
}

fileerror TDCTextDocument::PutDivider()
{
	if(FOut->GetLastChar() != '\x0a')
		PhysicalLineBreak();
	char temp[255] = "";
	for(int i=0;i<FLineLength;i++)
	{
		temp[strlen(temp)+1] = '\0';
		temp[strlen(temp)] = FDividerChar;
	}
	PutPhysical(temp);
	PhysicalLineBreak();
	FIsFirstLineOfDoc = FALSE;
	return FOut->GetResult();
}

fileerror TDCTextDocument::PutLink(char *LinkText, char *LinkTo)
{
	return PutText(LinkText);
}

fileerror TDCTextDocument::ParagraphBreak()
{
	if(FOut->GetLastChar() != '\x0a')
		PhysicalLineBreak();
	FIsFirstLineOfDoc = FALSE;
	PhysicalLineBreak();
	if(FStyleSheet == SS_FIRSTLINE)
		PutText("    ");
	return FOut->GetResult();
}

fileerror TDCTextDocument::LineBreak()
{
	FIsFirstLineOfDoc = FALSE;
	return PhysicalLineBreak();
}

fileerror TDCTextDocument::Tab()
{
	return PutPhysical("\t");
}

fileerror TDCTextDocument::PageBreak()
{
	return PutPhysical("\x0c");
}

fileerror TDCTextDocument::NewTopic(char *TopicTitle, char *TopicName)
{
	ParagraphBreak();
	PutDivider();
	SetStyleSheet(SS_HEADING1);
	PutText(TopicTitle);
	SetStyleSheet(SS_NORMAL);
	return FOut->GetResult();
}

int TDCTextDocument::GetIdent()
{
	return DT_TEXT;
}

TDCTextDocument::TDCTextDocument(char *Filename, TOutput *OutputObj) :
	TDCDocument(Filename, OutputObj)
{
	strcpy(FExtension,"TXT");
	FCharset = PCWINDOWS;
}

TDCTextDocument::~TDCTextDocument()
{

}